@@ -1,40 +1,11 @@
-# Copyright (C) 2002-2008, Stefan Schwarzer
-# All rights reserved.
-#
-# Redistribution and use in source and binary forms, with or without
-# modification, are permitted provided that the following conditions are
-# met:
-#
-# - Redistributions of source code must retain the above copyright
-#   notice, this list of conditions and the following disclaimer.
-#
-# - Redistributions in binary form must reproduce the above copyright
-#   notice, this list of conditions and the following disclaimer in the
-#   documentation and/or other materials provided with the distribution.
-#
-# - Neither the name of the above author nor the names of the
-#   contributors to the software may be used to endorse or promote
-#   products derived from this software without specific prior written
-#   permission.
-#
-# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
-# ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
-# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
-# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR
-# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
-# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
-# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
-# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
-# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
-# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
-# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
+# Copyright (C) 2002-2010, Stefan Schwarzer <sschwarzer@sschwarzer.net>
+# See the file LICENSE for licensing terms.
 
 """
 ftp_stat.py - stat result, parsers, and FTP stat'ing for `ftputil`
 """
 
-# $Id: ftp_stat.py 891 2009-10-30 19:37:10Z schwa $
-
+import math
 import re
 import stat
 import time
@@ -43,6 +14,10 @@ import ftp_error
 import ftp_stat_cache
 
 
+# These can be used to write custom parsers.
+__all__ = ['StatResult', 'Parser', 'UnixParser', 'MSParser']
+
+
 class StatResult(tuple):
     """
     Support class resembling a tuple like that returned from
@@ -58,13 +33,16 @@ class StatResult(tuple):
         # Don't call `__init__` via `super`. Construction from a
         #  sequence is implicitly handled by `tuple.__new__`, not
         #  `tuple.__init__`. As a by-product, this avoids a
-        #  `DeprecationWarning` in Python 2.6+
-        # these may be overwritten in a `Parser.parse_line` method
+        #  `DeprecationWarning` in Python 2.6+ .
+        # pylint: disable=W0231, W0613
+        #
+        # These may be overwritten in a `Parser.parse_line` method.
         self._st_name = ""
         self._st_target = None
+        self._st_mtime_precision = None
 
     def __getattr__(self, attr_name):
-        if self._index_mapping.has_key(attr_name):
+        if attr_name in self._index_mapping:
             return self[self._index_mapping[attr_name]]
         else:
             raise AttributeError("'StatResult' object has no attribute '%s'" %
@@ -79,7 +57,7 @@ class Parser(object):
     directory formats inherit from this class.
     """
 
-    # map month abbreviations to month numbers
+    # Map month abbreviations to month numbers.
     _month_numbers = {
       'jan':  1, 'feb':  2, 'mar':  3, 'apr':  4,
       'may':  5, 'jun':  6, 'jul':  7, 'aug':  8,
@@ -97,7 +75,11 @@ class Parser(object):
         If the line should be used to extract stat data from it,
         return a false value.
         """
-        # either a `_SRE_Match` instance or `None`
+        # Ignore empty lines stemming from only a line break.
+        if not line.strip():
+            # Yes, ignore the line if it's empty.
+            return True
+        # Either a `_SRE_Match` instance or `None`
         match = self._total_regex.search(line)
         return bool(match)
 
@@ -116,7 +98,7 @@ class Parser(object):
         raise NotImplementedError("must be defined by subclass")
 
     #
-    # helper methods for parts of a directory listing line
+    # Helper methods for parts of a directory listing line
     #
     def parse_unix_mode(self, mode_string):
         """
@@ -127,10 +109,11 @@ class Parser(object):
         If the mode string can't be parsed, raise an
         `ftp_error.ParserError`.
         """
-        st_mode = 0
         if len(mode_string) != 10:
             raise ftp_error.ParserError("invalid mode string '%s'" %
                                         mode_string)
+        st_mode = 0
+        #TODO Add support for "S" and sticky bit ("t", "T").
         for bit in mode_string[1:10]:
             bit = (bit != '-')
             st_mode = (st_mode << 1) + bit
@@ -138,8 +121,15 @@ class Parser(object):
             st_mode = st_mode | stat.S_ISUID
         if mode_string[6] == 's':
             st_mode = st_mode | stat.S_ISGID
-        file_type_to_mode = {'d': stat.S_IFDIR, 'l': stat.S_IFLNK,
-                             'c': stat.S_IFCHR, '-': stat.S_IFREG}
+        file_type_to_mode = {'b': stat.S_IFBLK, 'c': stat.S_IFCHR,
+                             'd': stat.S_IFDIR, 'l': stat.S_IFLNK,
+                             'p': stat.S_IFIFO, 's': stat.S_IFSOCK,
+                             '-': stat.S_IFREG,
+                             # Ignore types which `ls` can't make sense of
+                             #  (assuming the FTP server returns listings
+                             #  like `ls` does).
+                             '?': 0,
+                            }
         file_type = mode_string[0]
         if file_type in file_type_to_mode:
             st_mode = st_mode | file_type_to_mode[file_type]
@@ -149,7 +139,7 @@ class Parser(object):
         return st_mode
 
     def parse_unix_time(self, month_abbreviation, day, year_or_time,
-                        time_shift):
+                        time_shift, with_precision=False):
         """
         Return a floating point number, like from `time.mktime`, by
         parsing the string arguments `month_abbreviation`, `day` and
@@ -157,6 +147,14 @@ class Parser(object):
         "time on server" - "time on client" and is available as the
         `time_shift` parameter in the `parse_line` interface.
 
+        If `with_precision` is true (default: false), return a
+        two-element tuple consisting of the floating point number as
+        described in the previous paragraph and the precision of the
+        time in seconds. This takes into account that, for example, a
+        time string like "May 26  2005" has only a precision of one
+        day. This information is important for the `upload_if_newer`
+        and `download_if_newer` methods in the `FTPHost` class.
+
         Times in Unix-style directory listings typically have one of
         these formats:
 
@@ -173,34 +171,41 @@ class Parser(object):
             raise ftp_error.ParserError("invalid month name '%s'" % month)
         day = int(day)
         if ":" not in year_or_time:
-            # `year_or_time` is really a year
+            # `year_or_time` is really a year.
             year, hour, minute = int(year_or_time), 0, 0
             st_mtime = time.mktime( (year, month, day,
                                      hour, minute, 0, 0, 0, -1) )
+            # Precise up to a day.
+            st_mtime_precision = 24 * 60 * 60
         else:
-            # `year_or_time` is a time hh:mm
+            # `year_or_time` is a time hh:mm.
             hour, minute = year_or_time.split(':')
             year, hour, minute = None, int(hour), int(minute)
-            # try the current year
+            # Try the current year
             year = time.localtime()[0]
             st_mtime = time.mktime( (year, month, day,
                                      hour, minute, 0, 0, 0, -1) )
-            # rhs of comparison: transform client time to server time
+            # Times are precise up to a minute.
+            st_mtime_precision = 60
+            # Rhs of comparison: Transform client time to server time
             #  (as on the lhs), so both can be compared with respect
             #  to the set time shift (see the definition of the time
-            #  shift in `FTPHost.set_time_shift`'s docstring); the
+            #  shift in `FTPHost.set_time_shift`'s docstring). The
             #  last addend allows for small deviations between the
-            #  supposed (rounded) and the actual time shift
-            # #XXX the downside of this "correction" is that there is
+            #  supposed (rounded) and the actual time shift.
+            # #XXX The downside of this "correction" is that there is
             #  a one-minute time interval exactly one year ago that
             #  may cause that datetime to be recognized as the current
             #  datetime, but after all the datetime from the server
-            #  can only be exact up to a minute
-            if st_mtime > time.time() + time_shift + 60.0:
-                # if it's in the future, use previous year
+            #  can only be exact up to a minute.
+            if st_mtime > time.time() + time_shift + st_mtime_precision:
+                # If it's in the future, use previous year.
                 st_mtime = time.mktime( (year-1, month, day,
                                          hour, minute, 0, 0, 0, -1) )
-        return st_mtime
+        if with_precision:
+            return (st_mtime, st_mtime_precision)
+        else:
+            return st_mtime
 
     def parse_ms_time(self, date, time_, time_shift):
         """
@@ -220,8 +225,14 @@ class Parser(object):
         If this method can not make sense of the given arguments, it
         raises an `ftp_error.ParserError`.
         """
-        # don't complain about unused `time_shift` argument
-        # pylint: disable-msg=W0613
+        # Don't complain about unused `time_shift` argument
+        # pylint: disable=W0613
+        # For the time being, I don't add a `with_precision`
+        #  parameter as in the Unix parser because the precision for
+        #  the DOS format is always a minute and can be set in
+        #  `MSParser.parse_line`. Should you find yourself needing
+        #  support for `with_precision` for a derived class, please
+        #  send a mail (see ftputil.txt/html).
         try:
             month, day, year = [int(part) for part in date.split('-')]
             if year >= 70:
@@ -232,9 +243,9 @@ class Parser(object):
             hour, minute = int(hour), int(minute)
         except (ValueError, IndexError):
             raise ftp_error.ParserError("invalid time string '%s'" % time_)
-        if am_pm == 'A' and hour == 12:
+        if hour == 12 and am_pm == 'A':
             hour = 0
-        if am_pm == 'P' and hour != 12:
+        if hour != 12 and am_pm == 'P':
             hour = hour + 12
         st_mtime = time.mktime( (year, month, day,
                                  hour, minute, 0, 0, 0, -1) )
@@ -248,31 +259,31 @@ class UnixParser(Parser):
         """
         Split a line in metadata, nlink, user, group, size, month,
         day, year_or_time and name and return the result as an
-        nine-element list of these values.
+        nine-element list of these values. If the name is a link,
+        it will be encoded as a string "link_name -> link_target".
         """
         # This method encapsulates the recognition of an unusual
         #  Unix format variant (see ticket
-        #  http://ftputil.sschwarzer.net/trac/ticket/12 )
-        parts = line.split(None, 8)
-        if len(parts) == 9:
-            if parts[-1].startswith("-> "):
-                # for the alternative format, the last part will not be
-                #  "link_name -> link_target" but "-> link_target" and the
-                #  link name will be in the previous field;
-                # this heuristic will fail for names starting with "-> "
-                #  which should be _quite_ rare
-                # insert `None` for the user field
-                parts.insert(2, None)
-                parts[-2] = "%s %s" % tuple(parts[-2:])
-                del parts[-1]
-            return parts
-        elif len(parts) == 8:
-            # alternative unusual format, insert `None` for the user field
-            parts.insert(2, None)
-            return parts
-        else:
-            # no known Unix-style format
+        #  http://ftputil.sschwarzer.net/trac/ticket/12 ).
+        line_parts = line.split()
+        FIELD_COUNT_WITHOUT_USERID = 8
+        FIELD_COUNT_WITH_USERID = FIELD_COUNT_WITHOUT_USERID + 1
+        if len(line_parts) < FIELD_COUNT_WITHOUT_USERID:
+            # No known Unix-style format
             raise ftp_error.ParserError("line '%s' can't be parsed" % line)
+        # If we have a valid format (either with or without user id field),
+        #  the field with index 5 is either the month abbreviation or a day.
+        try:
+            int(line_parts[5])
+        except ValueError:
+            # Month abbreviation, "invalid literal for int"
+            line_parts = line.split(None, FIELD_COUNT_WITH_USERID-1)
+        else:
+            # Day
+            line_parts = line.split(None, FIELD_COUNT_WITHOUT_USERID-1)
+            USER_FIELD_INDEX = 2
+            line_parts.insert(USER_FIELD_INDEX, None)
+        return line_parts
 
     def parse_line(self, line, time_shift=0.0):
         """
@@ -295,17 +306,25 @@ class UnixParser(Parser):
         st_size = int(size)
         st_atime = None
         # st_mtime
-        st_mtime = self.parse_unix_time(month, day, year_or_time, time_shift)
+        st_mtime, st_mtime_precision = \
+          self.parse_unix_time(month, day, year_or_time, time_shift,
+                               with_precision=True)
         # st_ctime
         st_ctime = None
         # st_name
-        if " -> " in name:
+        if name.count(" -> ") > 1:
+            # If we have more than one arrow we can't tell where the link
+            #  name ends and the target name starts.
+            raise ftp_error.ParserError(
+                  'name "%s" contains more than one "->"' % name)
+        elif name.count(" -> ") == 1:
             st_name, st_target = name.split(' -> ')
         else:
             st_name, st_target = name, None
         stat_result = StatResult(
                       (st_mode, st_ino, st_dev, st_nlink, st_uid,
                        st_gid, st_size, st_atime, st_mtime, st_ctime) )
+        stat_result._st_mtime_precision = st_mtime_precision
         stat_result._st_name = st_name
         stat_result._st_target = st_target
         return stat_result
@@ -331,7 +350,7 @@ class MSParser(Parser):
             # "unpack list of wrong size"
             raise ftp_error.ParserError("line '%s' can't be parsed" % line )
         # st_mode
-        #  default to read access only; in fact, we can't tell
+        #  Default to read access only; in fact, we can't tell.
         st_mode = 0400
         if dir_or_size == "<DIR>":
             st_mode = st_mode | stat.S_IFDIR
@@ -363,6 +382,8 @@ class MSParser(Parser):
         # _st_name and _st_target
         stat_result._st_name = name
         stat_result._st_target = None
+        # mtime precision in seconds
+        stat_result._st_mtime_precision = 60
         return stat_result
 
 #
@@ -374,12 +395,12 @@ class _Stat(object):
     def __init__(self, host):
         self._host = host
         self._path = host.path
-        # use the Unix directory parser by default
+        # Use the Unix directory parser by default.
         self._parser = UnixParser()
-        # allow one chance to switch to another parser if the default
-        #  doesn't work
+        # Allow one chance to switch to another parser if the default
+        #  doesn't work.
         self._allow_parser_switching = True
-        # cache only lstat results; `stat` works locally on `lstat` results
+        # Cache only lstat results. `stat` works locally on `lstat` results.
         self._lstat_cache = ftp_stat_cache.StatCache()
 
     def _host_dir(self, path):
@@ -389,6 +410,36 @@ class _Stat(object):
         """
         return self._host._dir(path)
 
+    def _stat_results_from_dir(self, path):
+        """
+        Yield stat results extracted from the directory listing `path`.
+        Omit the special entries for the directory itself and its parent
+        directory.
+        """
+        lines = self._host_dir(path)
+        # `cache` is the "high-level" `StatCache` object whereas
+        #  `cache._cache` is the "low-level" `LRUCache` object.
+        cache = self._lstat_cache
+        # Auto-grow cache if the cache up to now can't hold as many
+        #  entries as there are in the directory `path`.
+        if cache._enabled and len(lines) >= cache._cache.size:
+            new_size = int(math.ceil(1.1 * len(lines)))
+            cache.resize(new_size)
+        # Yield stat results from lines.
+        for line in lines:
+            if self._parser.ignores_line(line):
+                continue
+            # For `listdir`, we are interested in just the names,
+            #  but we use the `time_shift` parameter to have the
+            #  correct timestamp values in the cache.
+            stat_result = self._parser.parse_line(line,
+                                                  self._host.time_shift())
+            if stat_result._st_name in [self._host.curdir, self._host.pardir]:
+                continue
+            loop_path = self._path.join(path, stat_result._st_name)
+            self._lstat_cache[loop_path] = stat_result
+            yield stat_result
+
     def _real_listdir(self, path):
         """
         Return a list of directories, files etc. in the directory
@@ -397,32 +448,18 @@ class _Stat(object):
         If the directory listing from the server can't be parsed
         raise a `ParserError`.
         """
-        # we _can't_ put this check into `FTPHost._dir`; see its docstring
+        # We _can't_ put this check into `FTPHost._dir`; see its docstring.
         path = self._path.abspath(path)
-        # `listdir` should only be allowed for directories and links to them
+        # `listdir` should only be allowed for directories and links to them.
         if not self._path.isdir(path):
             raise ftp_error.PermanentError(
                   "550 %s: no such directory or wrong directory parser used" %
                   path)
-        # set up for loop
-        lines = self._host_dir(path)
-        # exit the method now if there aren't any files
-        if lines == ['']:
-            return []
+        # Set up for `for` loop.
         names = []
-        for line in lines:
-            if self._parser.ignores_line(line):
-                continue
-            # for `listdir`, we are interested in just the names,
-            #  but we use the `time_shift` parameter to have the
-            #  correct timestamp values in the cache
-            stat_result = self._parser.parse_line(line,
-                                                  self._host.time_shift())
-            loop_path = self._path.join(path, stat_result._st_name)
-            self._lstat_cache[loop_path] = stat_result
+        for stat_result in self._stat_results_from_dir(path):
             st_name = stat_result._st_name
-            if st_name not in (self._host.curdir, self._host.pardir):
-                names.append(st_name)
+            names.append(st_name)
         return names
 
     def _real_lstat(self, path, _exception_for_missing_path=True):
@@ -440,7 +477,7 @@ class _Stat(object):
         _not_ intended for use by ftputil clients.)
         """
         path = self._path.abspath(path)
-        # if the path is in the cache, return the lstat result
+        # If the path is in the cache, return the lstat result.
         if path in self._lstat_cache:
             return self._lstat_cache[path]
         # Note: (l)stat works by going one directory up and parsing
@@ -450,7 +487,6 @@ class _Stat(object):
             raise ftp_error.RootDirError(
                   "can't stat remote root directory")
         dirname, basename = self._path.split(path)
-
         # If even the directory doesn't exist and we don't want the
         #  exception, treat it the same as if the path wasn't found in
         #  the directory's contents (compare below). The use of `isdir`
@@ -458,37 +494,30 @@ class _Stat(object):
         #  will at the latest stop when we've got to the root directory.
 #         if not self._path.isdir(dirname) and not _exception_for_missing_path:
 #             return None
-        # loop through all lines of the directory listing; we
+        # Loop through all lines of the directory listing. We
         #  probably won't need all lines for the particular path but
         #  we want to collect as many stat results in the cache as
-        #  possible
+        #  possible.
         lstat_result_for_path = None
-        lines = self._host_dir(dirname)
-        for line in lines:
-            if self._parser.ignores_line(line):
-                continue
-            stat_result = self._parser.parse_line(line,
-                          self._host.time_shift())
-            loop_path = self._path.join(dirname, stat_result._st_name)
-            self._lstat_cache[loop_path] = stat_result
-            # needed to work without cache or with disabled cache
+        for stat_result in self._stat_results_from_dir(dirname):
+            # Needed to work without cache or with disabled cache.
             if stat_result._st_name == basename:
                 lstat_result_for_path = stat_result
         if lstat_result_for_path is not None:
             return lstat_result_for_path
-        # path was not found during the loop
+        # Path was not found during the loop.
         if _exception_for_missing_path:
-            #TODO use FTP DIR command on the file to implicitly use
+            #TODO Use FTP DIR command on the file to implicitly use
             #  the usual status code of the server for missing files
-            #  (450 vs. 550)
+            #  (450 vs. 550).
             raise ftp_error.PermanentError(
                   "550 %s: no such file or directory" % path)
         else:
-            # be explicit; returning `None` is a signal for
+            # Be explicit. Returning `None` is a signal for
             #  `_Path.exists/isfile/isdir/islink` that the path was
-            #  not found; if we would raise an exception, there would
+            #  not found. If we would raise an exception, there would
             #  be no distinction between a missing path or a more
-            #  severe error in the code above
+            #  severe error in the code above.
             return None
 
     def _real_stat(self, path, _exception_for_missing_path=True):
@@ -504,35 +533,35 @@ class _Stat(object):
         (`_exception_for_missing_path` is an implementation aid and
         _not_ intended for use by ftputil clients.)
         """
-        # save for error message
+        # Save for error message.
         original_path = path
-        # most code in this method is used to detect recursive
-        #  link structures
-        visited_paths = {}
+        # Most code in this method is used to detect recursive
+        #  link structures.
+        visited_paths = set()
         while True:
-            # stat the link if it is one, else the file/directory
+            # Stat the link if it is one, else the file/directory.
             lstat_result = self._real_lstat(path, _exception_for_missing_path)
             if lstat_result is None:
                 return None
-            # if the file is not a link, the `stat` result is the
-            #  same as the `lstat` result
+            # If the file is not a link, the `stat` result is the
+            #  same as the `lstat` result.
             if not stat.S_ISLNK(lstat_result.st_mode):
                 return lstat_result
-            # if we stat'ed a link, calculate a normalized path for
-            #  the file the link points to
-            # we don't use `basename`
-            # pylint: disable-msg=W0612
+            # If we stat'ed a link, calculate a normalized path for
+            #  the file the link points to.
+            # We don't use `basename`.
+            # pylint: disable=W0612
             dirname, basename = self._path.split(path)
             path = self._path.join(dirname, lstat_result._st_target)
             path = self._path.abspath(self._path.normpath(path))
-            # check for cyclic structure
+            # Check for cyclic structure.
             if path in visited_paths:
-                # we had this path already
+                # We had seen this path already.
                 raise ftp_error.PermanentError(
                       "recursive link structure detected for remote path '%s'" %
                       original_path)
-            # remember the path we have encountered
-            visited_paths[path] = True
+            # Remember the path we have encountered.
+            visited_paths.add(path)
 
     def __call_with_parser_retry(self, method, *args, **kwargs):
         """
@@ -547,8 +576,8 @@ class _Stat(object):
         #  parser - which is wrong.
         try:
             result = method(*args, **kwargs)
-            # if a `listdir` call didn't find anything, we can't
-            #  say anything about the usefulness of the parser
+            # If a `listdir` call didn't find anything, we can't
+            #  say anything about the usefulness of the parser.
             if (method is not self._real_listdir) and result:
                 self._allow_parser_switching = False
             return result
@@ -560,7 +589,9 @@ class _Stat(object):
             else:
                 raise
 
-    def listdir(self, path):
+    # Don't use these methods, but instead the corresponding methods
+    #  in the `FTPHost` class.
+    def _listdir(self, path):
         """
         Return a list of items in `path`.
 
@@ -570,7 +601,7 @@ class _Stat(object):
         """
         return self.__call_with_parser_retry(self._real_listdir, path)
 
-    def lstat(self, path, _exception_for_missing_path=True):
+    def _lstat(self, path, _exception_for_missing_path=True):
         """
         Return a `StatResult` without following links.
 
@@ -581,7 +612,7 @@ class _Stat(object):
         return self.__call_with_parser_retry(self._real_lstat, path,
                                              _exception_for_missing_path)
 
-    def stat(self, path, _exception_for_missing_path=True):
+    def _stat(self, path, _exception_for_missing_path=True):
         """
         Return a `StatResult` with following links.
 
